(* FOLDER.PAS - Program to insert a message into a specified folder. 
   Warning: SYS$SCRATCH must be defined to point at a writable disk
   directory for proper operation. MM and MINT support has been removed
   from the copy for public distribution. *)

[inherit ('SYS$LIBRARY:STARLET')] program folder;

const  ALFA_SIZE = 252;
       BIGALFA_SIZE = 1024;

type

     (* Generic VMS stuff. *)

     $BYTE = [BYTE] -128..127;
     $WORD = [WORD] -32768..32767;
     $QUAD = [QUAD, UNSAFE] RECORD
                              L0 : UNSIGNED;
                              L1 : INTEGER;
                              END;
     $OCTA = [OCTA, UNSAFE] RECORD
                              L0, L1, L2 : UNSIGNED;
                              L3 : INTEGER;
                              END;
     $UBYTE = [BYTE] 0..255;
     $UWORD = [WORD] 0..65535;
     $UQUAD = [QUAD, UNSAFE] RECORD
                               L0, L1 : UNSIGNED;
                               END;
     $UOCTA = [OCTA, UNSAFE] RECORD
                               L0, L1, L2, L3 : UNSIGNED;
                               END;
     $QWAD = [QUAD, UNSAFE] ARRAY [1..4] OF $WORD;

        itemtype = [byte (12)] record
          buffersize          : [pos (00)] $uword;
          itemcode            : [pos (16)] $uword;
          bufferaddress       : [pos (32), unsafe] integer;
          returnlengthaddress : [pos (64), unsafe] integer;
        end; (* itemtype record *)
        itemlist = array [1..10] of itemtype;

        string = varying [ALFA_SIZE] of char;
        bigstring = varying [BIGALFA_SIZE] of char;
        vstring = string;
        bigvstring = bigstring;

        file_list_ptr = ^file_list_item;
        file_list_item = record
                           next : file_list_ptr;
                           name : string;
                         end; (* file_list_item record *)

var
        inputitems, outputitems : [static] itemlist;
        mailfilecontext, messagecontext, pid : [static] unsigned;
        mail_file_default, mail_file_name, folder_name,
          temp_file_name, from_string, to_string, cc_string,
          subj_string : [static] string;
        stat, longlength, i : integer;
        time_string : packed array [1..23] of char;
        temp_file, message_file : text;
        buffer : bigvstring;
        user_name, tmpv, envelope_from, message_file_name,
          header_file_name : vstring;
        result_length : $uword;
        file_status : (starting, moving, processing, finished);
        file_errors, file_found, blank_spacers : boolean;
        file_list_start, file_list_next : file_list_ptr;
        process_privileges,
          current_privileges, image_privileges : [static] $uquad;
        privileges_on : [static] boolean := false;

  (* Error message codes defined in FOLDER.MSG *)

  FOLDER__CANOPNMSG, FOLDER__CANOPNTMP, FOLDER__ERRGETJPI, FOLDER__ERRPRIV,
  FOLDER__ERRINMAIL, FOLDER__CANTRESET, FOLDER__CANTREADHEADER,
  FOLDER__SENDABORT : [external, value] integer;

[unbound, external] function CLI$GET_VALUE (
  name : [class_s] packed array [l1..u1 : integer] of char;
  var result : [class_s] packed array [l2..u2 : integer] of char;
  var result_length : $uword := %IMMED 0) : integer; extern;

[unbound, external] function CLI$PRESENT (
  name : [class_s] packed array [l1..u1 : integer] of char) : integer; extern;

[unbound, external] function MAIL$MAILFILE_BEGIN (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MAILFILE_OPEN (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MAILFILE_CLOSE (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MAILFILE_END (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MESSAGE_BEGIN (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MESSAGE_COPY (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$MESSAGE_END (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_ABORT (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_BEGIN (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_ADD_ADDRESS (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_ADD_ATTRIBUTE (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_ADD_BODYPART (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_MESSAGE (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] function MAIL$SEND_END (
  var context : unsigned; var in_list, out_list : itemlist) : integer; extern;

[unbound, external] procedure LIB$SIGNAL (
  %IMMED stat : [list, unsafe] integer); extern;

procedure enable_privileges;

begin (* enable_privileges *)
  if not privileges_on then begin
    $SETPRV (ENBFLG := 1, PRVADR := image_privileges);
    privileges_on := true;
  end;
end; (* enable_privileges *)

procedure disable_privileges;

begin (* disable_privileges *)
  if privileges_on then begin
    $SETPRV (ENBFLG := 0, PRVADR := image_privileges);
    privileges_on := false;
  end;
end; (* disable_privileges *)

function qu_rtxt (var line : bigvstring) : boolean;

label
  10;

begin (* qu_rtxt *)
10:
  line.length := 0;
  case file_status of
    finished :
      qu_rtxt := false;
    starting, moving :
      if file_list_next <> nil then begin
        open (file_variable := message_file,
              file_name := file_list_next^.name,
              history := READONLY, error := CONTINUE,
              organization := SEQUENTIAL,
              sharing := READWRITE, default := '.TXT');
        if status (message_file) <= 0 then
          reset (file_variable := message_file, error := CONTINUE);
        if status (message_file) > 0 then begin
          LIB$SIGNAL (FOLDER__CANOPNMSG, 2, file_list_next^.name.length,
                      iaddress (file_list_next^.name.body));
          close (file_variable := message_file, error := CONTINUE);
          file_list_next := file_list_next^.next;
          file_errors := true;
          goto 10;
        end; (* if *)
        file_found := true;
        if file_status = starting then begin
          file_status := processing;
          goto 10;
        end; (* if *)
        file_status := processing;
        if not blank_spacers then goto 10;
        line := line + chr (10);
        qu_rtxt := true;
      end else begin
        file_status := finished;
        qu_rtxt := false;
      end; (* if *)
    processing :
      if not eof (message_file) then begin
        readln (message_file, line);
        if line.length >= BIGALFA_SIZE then line.length := pred (BIGALFA_SIZE);
        line := line + chr (10);
        qu_rtxt := true;
      end else begin
        close (file_variable := message_file, error := CONTINUE);
        file_list_next := file_list_next^.next;
        file_status := moving;
        goto 10;
      end;
  end; (* case *)
end; (* qu_rtxt *)

begin (* folder *)
  user_name.length := 0;
  inputitems[1].itemcode := JPI$_USERNAME;
  inputitems[1].bufferaddress := iaddress (user_name.body);
  inputitems[1].buffersize := ALFA_SIZE;
  inputitems[1].returnlengthaddress := iaddress (user_name.length);
  inputitems[2].itemcode := JPI$_PID;
  inputitems[2].bufferaddress := iaddress (pid);
  inputitems[2].buffersize := 4;
  inputitems[2].returnlengthaddress := 0;
  inputitems[3].itemcode := JPI$_PROCPRIV;
  inputitems[3].bufferaddress := iaddress (process_privileges);
  inputitems[3].buffersize := 8;
  inputitems[3].returnlengthaddress := 0;
  inputitems[4].itemcode := JPI$_IMAGPRIV;
  inputitems[4].bufferaddress := iaddress (image_privileges);
  inputitems[4].buffersize := 8;
  inputitems[4].returnlengthaddress := 0;
  inputitems[5] := zero;
  stat := $GETJPIW ( , , , inputitems);
  if not odd (stat) then $EXIT (FOLDER__ERRGETJPI);
  
  image_privileges.l0 := UAND (image_privileges.l0,
                               UNOT (process_privileges.l0));
  image_privileges.l1 := UAND (image_privileges.l1,
                               UNOT (process_privileges.l1));

  stat := $SETPRV (ENBFLG := 0, PRVADR := image_privileges);
  if not odd (stat) then LIB$SIGNAL (FOLDER__ERRPRIV, 1, stat);

  blank_spacers := odd (CLI$PRESENT ('BLANK'));

  file_list_start := nil;
  while odd (CLI$GET_VALUE ('MESSAGEFILE', temp_file_name.body,
                            temp_file_name.length)) do begin
    if file_list_start = nil then begin
      new (file_list_start);
      file_list_next := file_list_start;
    end else begin
      new (file_list_next^.next);
      file_list_next := file_list_next^.next;
    end; (* if *)
    file_list_next^.name := temp_file_name;
    file_list_next^.next := nil;
  end; (* while *)
  file_list_next := file_list_start;

  stat := CLI$PRESENT ('MM');
  if not odd (stat) then begin

    stat := CLI$PRESENT ('MINT');
    if not odd (stat) then begin

      stat := CLI$GET_VALUE ('FROM', from_string.body, from_string.length);
      if not odd (stat) then
        from_string := substr (user_name.body, 1, user_name.length);

      stat := CLI$GET_VALUE ('FOLDER', folder_name.body, folder_name.length);
      if not odd (stat) then folder_name := 'NEWMAIL';

      stat := CLI$GET_VALUE ('MAILFILE', mail_file_name.body,
                             mail_file_name.length);
      if not odd (stat) then mail_file_name := 'MAIL.MAI';

      stat := CLI$GET_VALUE ('TO', to_string.body, to_string.length);
      if not odd (stat) then to_string :=
        substr (user_name.body, 1, user_name.length);
      stat := CLI$GET_VALUE ('CC', cc_string.body, cc_string.length);
      if not odd (stat) then cc_string := '';
      stat := CLI$GET_VALUE ('SUBJECT', subj_string.body, subj_string.length);
      if not odd (stat) then subj_string := '';

      if folder_name = 'NEWMAIL' then begin

        inputitems[1] := zero;
        outputitems[1] := zero;
        stat := MAIL$SEND_BEGIN (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        inputitems[1].buffersize := from_string.length;
        inputitems[1].itemcode := MAIL$_SEND_FROM_LINE;
        inputitems[1].bufferaddress := iaddress (from_string.body);
        inputitems[1].returnlengthaddress := 0;
        inputitems[2].buffersize := to_string.length;
        inputitems[2].itemcode := MAIL$_SEND_TO_LINE;
        inputitems[2].bufferaddress := iaddress (to_string.body);
        inputitems[2].returnlengthaddress := 0;
        inputitems[3].buffersize := cc_string.length;
        inputitems[3].itemcode := MAIL$_SEND_CC_LINE;
        inputitems[3].bufferaddress := iaddress (cc_string.body);
        inputitems[3].returnlengthaddress := 0;
        inputitems[4].buffersize := subj_string.length;
        inputitems[4].itemcode := MAIL$_SEND_SUBJECT;
        inputitems[4].bufferaddress := iaddress (subj_string.body);
        inputitems[4].returnlengthaddress := 0;
        inputitems[5] := zero;
        enable_privileges;
        stat := MAIL$SEND_ADD_ATTRIBUTE (mailfilecontext, inputitems,
                                         outputitems);
        disable_privileges;
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        temp_file_name := '_' + substr (user_name.body, 1, user_name.length);
        inputitems[1].buffersize := temp_file_name.length;
        inputitems[1].itemcode := MAIL$_SEND_USERNAME;
        inputitems[1].bufferaddress := iaddress (temp_file_name.body);
        inputitems[1].returnlengthaddress := 0;
        inputitems[2] := zero;
        stat := MAIL$SEND_ADD_ADDRESS (mailfilecontext,
                                       inputitems, outputitems);
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        inputitems[1].itemcode := MAIL$_SEND_RECORD;
        inputitems[1].bufferaddress := iaddress (buffer.body);
        inputitems[1].returnlengthaddress := 0;
        while qu_rtxt (buffer) do begin
          inputitems[1].buffersize := buffer.length - 1;
          stat := MAIL$SEND_ADD_BODYPART (mailfilecontext, inputitems,
                                          outputitems);
          if not odd (stat) then begin
            close (message_file, error := CONTINUE);
            LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
          end; (* if *)
        end; (* while *)
        if file_errors and (not file_found) then begin
          inputitems[1] := zero;
          stat := MAIL$SEND_ABORT (mailfilecontext, inputitems, outputitems);
          LIB$SIGNAL (FOLDER__SENDABORT);
        end; (* if *)
        inputitems[1] := zero;
        stat := MAIL$SEND_MESSAGE (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        stat := MAIL$SEND_END (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);

      end else begin

        stat := $ASCTIM (timbuf := time_string);
        if not odd (stat) then $EXIT (stat);
        writev (temp_file_name, 'SYS$SCRATCH:FOLDER_', pid:0,
                time_string[13], time_string[14], time_string[16], 
                time_string[17], time_string[19], time_string[20], '.TMP;1');

        from_string := from_string + ' ' + time_string;

        open (file_variable := temp_file, file_name := temp_file_name,
              record_length := 1024, record_type := VARIABLE,
              organization := SEQUENTIAL, sharing := NONE,
              error := CONTINUE, history := NEW);
        if status (temp_file) <= 0 then
          rewrite (file_variable := temp_file, error := CONTINUE);
        if status (temp_file) > 0 then
          LIB$SIGNAL (FOLDER__CANOPNTMP, 2, temp_file_name.length,
                      iaddress (temp_file_name.body));

        writeln (temp_file, chr (12));
        writeln (temp_file, 'From:', chr (9), from_string);
        writeln (temp_file, 'To:', chr (9), to_string);
        writeln (temp_file, 'CC:', chr (9), cc_string);
        writeln (temp_file, 'Subj:', chr (9), subj_string);
        writeln (temp_file);

        while qu_rtxt (buffer) do
          writeln (temp_file, substr (buffer.body, 1, buffer.length - 1));

        if file_errors and (not file_found) then begin
          close (temp_file, disposition := DELETE);
          LIB$SIGNAL (FOLDER__SENDABORT);
        end; (* if *)

        close (temp_file);

        inputitems[1] := zero;
        outputitems[1].buffersize := ALFA_SIZE;
        outputitems[1].itemcode := MAIL$_MAILFILE_MAIL_DIRECTORY;
        outputitems[1].bufferaddress := iaddress (mail_file_default.body);
        outputitems[1].returnlengthaddress := iaddress (longlength);
        outputitems[2] := zero;
        mailfilecontext := 0;
        stat := MAIL$MAILFILE_BEGIN (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        mail_file_default.length := longlength;
        mail_file_default := mail_file_default + 'MAIL.MAI;';
        inputitems[1].buffersize := temp_file_name.length;
        inputitems[1].itemcode := MAIL$_MAILFILE_NAME;
        inputitems[1].bufferaddress := iaddress (temp_file_name.body);
        inputitems[1].returnlengthaddress := 0;
        inputitems[2] := zero;
        outputitems[1] := zero;
        stat := MAIL$MAILFILE_OPEN (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        inputitems[1].buffersize := 4;
        inputitems[1].itemcode := MAIL$_MESSAGE_FILE_CTX;
        inputitems[1].bufferaddress := iaddress (mailfilecontext);;
        inputitems[1].returnlengthaddress := 0;
        messagecontext := 0;
        stat := MAIL$MESSAGE_BEGIN (messagecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        inputitems[1].buffersize := mail_file_default.length;
        inputitems[1].itemcode := MAIL$_MESSAGE_DEFAULT_NAME;
        inputitems[1].bufferaddress := iaddress (mail_file_default.body);;
        inputitems[1].returnlengthaddress := 0;
        inputitems[2].buffersize := mail_file_name.length;
        inputitems[2].itemcode := MAIL$_MESSAGE_FILENAME;
        inputitems[2].bufferaddress := iaddress (mail_file_name.body);;
        inputitems[2].returnlengthaddress := 0;
        inputitems[3].buffersize := 0;
        inputitems[3].itemcode := MAIL$_MESSAGE_NEXT;
        inputitems[3].bufferaddress := 0;
        inputitems[3].returnlengthaddress := 0;
        inputitems[4].buffersize := folder_name.length;
        inputitems[4].itemcode := MAIL$_MESSAGE_FOLDER;
        inputitems[4].bufferaddress := iaddress (folder_name.body);
        inputitems[4].returnlengthaddress := 0;
        inputitems[5] := zero;
        stat := MAIL$MESSAGE_COPY (messagecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        inputitems[1] := zero;
        outputitems[1] := zero;
        stat := MAIL$MESSAGE_END (messagecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        stat := MAIL$MAILFILE_CLOSE (mailfilecontext, inputitems, outputitems);
        if not odd (stat) then begin
          delete_file (temp_file_name);
          if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
        end; (* if *)
        stat := MAIL$MAILFILE_END (mailfilecontext, inputitems, outputitems);
        delete_file (temp_file_name);
        if not odd (stat) then LIB$SIGNAL (FOLDER__ERRINMAIL, 0, stat);
      end; (* if *)

    end else begin

      (* MINT support goes here. *)

    end; (* if *)

  end else begin

    (* MM support goes here. *)

  end; (* if *)

  $EXIT (1);
end. (* folder *)
